#include <IotWebConf.h>
#include <IotWebConfTParameter.h>

/** Other scripts can use this namespace, in order to define commands, via callbacks.
 * Those callbacks can then be used to send commands to the sensor (reset, calibrate, led on/off, ...)
 * The callbacks can either have no parameter, one int32_t parameter or one char pointer.
 */

namespace sensor_console {
  void defineCommand(const char *name, void (*function)(), const __FlashStringHelper *doc_fstring);
  void defineIntCommand(const char *name, void (*function)(int32_t), const __FlashStringHelper *doc_fstring);
  void defineStringCommand(const char *name, void (*function)(char*), const __FlashStringHelper *doc_fstring);

  void checkSerialInput();

  void execute(const char *command_line);
}

// -- Initial name of the Thing. Used e.g. as SSID of the own Access Point.
const char thingName[] = "testThing";

// -- Initial password to connect to the Thing, when it creates an own Access Point.
const char wifiInitialApPassword[] = "smrtTHNG8266";

#define STRING_LEN 128
#define NUMBER_LEN 32

// -- Configuration specific key. The value should be modified if config structure was changed.
#define CONFIG_VERSION "dem99"

// -- Method declarations.
void handleRoot();
// -- Callback methods.
void configSaved();

DNSServer dnsServer;
WebServer server(80);

static const char chooserValues[][STRING_LEN] = { "red", "blue", "darkYellow" };
static const char chooserNames[][STRING_LEN] = { "Red", "Blue", "Dark yellow" };

IotWebConf iotWebConf(thingName, &dnsServer, &server, wifiInitialApPassword, CONFIG_VERSION);
iotwebconf::TextTParameter<STRING_LEN> stringParam =
  iotwebconf::Builder<iotwebconf::TextTParameter<STRING_LEN>>("stringParam").
  label("String param").
  defaultValue("").
  build();
iotwebconf::ParameterGroup group1 = iotwebconf::ParameterGroup("group1", "");
iotwebconf::IntTParameter<int16_t> intParam =
  iotwebconf::Builder<iotwebconf::IntTParameter<int16_t>>("intParam").
  label("Int param").
  defaultValue(30).
  min(1).
  max(100).
  step(1).
  placeholder("1..100").
  build();
// -- We can add a legend to the separator
iotwebconf::ParameterGroup group2 = iotwebconf::ParameterGroup("c_factor", "Calibration factor");
iotwebconf::FloatTParameter floatParam =
   iotwebconf::Builder<iotwebconf::FloatTParameter>("floatParam").
   label("Float param").
   defaultValue(0.0).
   step(0.1).
   placeholder("e.g. 23.4").
   build();
iotwebconf::CheckboxTParameter checkboxParam =
   iotwebconf::Builder<iotwebconf::CheckboxTParameter>("checkParam").
   label("Check param").
   defaultValue(true).
   build();
iotwebconf::SelectTParameter<STRING_LEN> chooserParam =
   iotwebconf::Builder<iotwebconf::SelectTParameter<STRING_LEN>>("chooseParam").
   label("Choose param").
   optionValues((const char*)chooserValues).
   optionNames((const char*)chooserNames).
   optionCount(sizeof(chooserValues) / STRING_LEN).
   nameLength(STRING_LEN).
   defaultValue("blue").
   build();
iotwebconf::ColorTParameter colorParam =
   iotwebconf::Builder<iotwebconf::ColorTParameter>("colorParam").
   label("Choose color").
   defaultValue("#FFDD88").
   build();
iotwebconf::DateTParameter dateParam =
   iotwebconf::Builder<iotwebconf::DateTParameter>("dateParam").
   label("Select date").
   defaultValue("").
   build();
iotwebconf::TimeTParameter timeParam =
   iotwebconf::Builder<iotwebconf::TimeTParameter>("timeParam").
   label("Select time").
   defaultValue("").
   build();

void setup() 
{
  Serial.begin(115200);
  Serial.println();
  Serial.println("Starting up...");

  group1.addItem(&intParam);
  group2.addItem(&floatParam);
  group2.addItem(&checkboxParam);
  group2.addItem(&chooserParam);
  group2.addItem(&colorParam);
  group2.addItem(&dateParam);
  group2.addItem(&timeParam);

  iotWebConf.addSystemParameter(&stringParam);
  iotWebConf.addParameterGroup(&group1);
  iotWebConf.addParameterGroup(&group2);
  iotWebConf.setConfigSavedCallback(&configSaved);
  iotWebConf.getApTimeoutParameter()->visible = true;

  // -- Initializing the configuration.
  iotWebConf.init();

  // -- Set up required URL handlers on the web server.
  server.on("/", handleRoot);
  server.on("/config", []{ iotWebConf.handleConfig(); });
  server.onNotFound([](){ iotWebConf.handleNotFound(); });
  
  sensor_console::defineCommand("reset", [](){ ESP.restart(); }, F("(Restarts the ESP)"));
  
  sensor_console::defineCommand("save_config", [](){ iotWebConf.saveConfig(); }, F("(Saves the config to EEPROM)"));

  sensor_console::defineCommand("reset_config", []() {
    Serial.println(F("Resetting config..."));
    iotWebConf.getRootParameterGroup()->applyDefaultValue();
    Serial.println(F("Done!"));
  }, F("(Resets the complete IotWeb config)"));

  sensor_console::defineStringCommand("ssid", [](char *ssid) {
    Serial.print(F("Setting WiFi ssid to "));
    Serial.println(ssid);
    strlcpy(iotWebConf.getWifiSsidParameter()->valueBuffer, ssid, iotWebConf.getWifiSsidParameter()->getLength());
  },   F("name (Sets SSID to name)"));

  sensor_console::defineStringCommand("wifi_pwd", [](char *pwd) {
    Serial.print(F("Setting WiFi password to "));
    Serial.println(pwd);
    strlcpy(iotWebConf.getWifiPasswordParameter()->valueBuffer, pwd, iotWebConf.getWifiPasswordParameter()->getLength());
  }, F("abc (Sets WiFi password to abc)"));
  
  sensor_console::defineStringCommand("ap_pwd", [](char *pwd) {
    Serial.print(F("Setting AP password to "));
    Serial.println(pwd);
    strlcpy(iotWebConf.getApPasswordParameter()->valueBuffer, pwd, iotWebConf.getApPasswordParameter()->getLength());
  }, F("abc (Sets AP password to abc)"));
  
  sensor_console::defineIntCommand("ap", [](int onOff) {
    if (onOff) {
      Serial.print(F("Enable "));
    } else {
      Serial.print(F("Disable "));
    }
    Serial.println(F("AP mode!"));
    iotWebConf.forceApMode(onOff);
  }, F("0/1 (Enables/disables access point)"));
  
  sensor_console::defineIntCommand("wifi", [](int onOff) {
    if (onOff) {
      Serial.print(F("Enable "));
      iotWebConf.goOnLine();
    } else {
      Serial.print(F("Disable "));
      iotWebConf.goOffLine();
    }
    Serial.println(F("Wifi!"));
  }, F("0/1 (Enables/disables WiFi)"));
  
  //TODO: Remove fluff
  //TODO: Add wifi on/off?

  Serial.println("Ready.");
}

void loop() 
{
  // -- doLoop should be called as frequently as possible.
  iotWebConf.doLoop();
  sensor_console::checkSerialInput();
}

/**
 * Handle web requests to "/" path.
 */
void handleRoot()
{
  // -- Let IotWebConf test and handle captive portal requests.
  if (iotWebConf.handleCaptivePortal())
  {
    // -- Captive portal request were already served.
    return;
  }
  String s = "<!DOCTYPE html><html lang=\"en\"><head><meta name=\"viewport\" content=\"width=device-width, initial-scale=1, user-scalable=no\"/>";
  s += "<title>IotWebConf 03 Custom Parameters</title></head><body>Hello world!";
  s += "<ul>";
  s += "<li>String param value: ";
  s += stringParam.value();
  s += "<li>Int param value: ";
  s += intParam.value();
  s += "<li>Float param value: ";
  s += floatParam.value();
  s += "<li>CheckBox selected: ";
  s += checkboxParam.isChecked();
  s += "<li>Option selected: ";
  s += chooserParam.value();
  s += "<li>Color selected: <div style='background-color:";
  s += colorParam.value();
  s += "';> sample </div> (";
  s += colorParam.value();
  s += ")";
  s += "<li>Date value: ";
  s += dateParam.value();
  s += "<li>Time value: ";
  s += timeParam.value();
  s += "</ul>";
  s += "Go to <a href='config'>configure page</a> to change values.";
  s += "</body></html>\n";

  server.send(200, "text/html", s);
}

void configSaved()
{
  Serial.println("Configuration was updated.");
}

/*** 
 * Sensor console
 */

namespace sensor_console {
  const uint8_t MAX_COMMANDS = 10;
  const uint8_t MAX_COMMAND_SIZE = 40;

  uint8_t commands_count = 0;

  enum input_type {
    NONE,
    INT32,
    STRING
  };

  struct Command {
    const char *name;
    union {
      void (*voidFunction)();
      void (*intFunction)(int32_t);
      void (*strFunction)(char*);
    };
    const char *doc;
    input_type parameter_type;
  };

  struct CommandLine {
    char function_name[MAX_COMMAND_SIZE];
    input_type argument_type;
    int32_t int_argument;
    char str_argument[MAX_COMMAND_SIZE];
  };

  Command commands[MAX_COMMANDS];

  bool addCommand(const char *name, const __FlashStringHelper *doc_fstring) {
    if (commands_count < MAX_COMMANDS) {
      commands[commands_count].name = name;
      commands[commands_count].doc = (const char*) doc_fstring;
      return true;
    } else {
      Serial.println(F("Too many commands have been defined."));
      return false;
    }
  }

  void defineCommand(const char *name, void (*function)(), const __FlashStringHelper *doc_fstring) {
    if (addCommand(name, doc_fstring)) {
      commands[commands_count].voidFunction = function;
      commands[commands_count++].parameter_type = NONE;
    }
  }

  void defineIntCommand(const char *name, void (*function)(int32_t), const __FlashStringHelper *doc_fstring) {
    if (addCommand(name, doc_fstring)) {
      commands[commands_count].intFunction = function;
      commands[commands_count++].parameter_type = INT32;
    }
  }

  void defineStringCommand(const char *name, void (*function)(char*), const __FlashStringHelper *doc_fstring) {
    if (addCommand(name, doc_fstring)) {
      commands[commands_count].strFunction = function;
      commands[commands_count++].parameter_type = STRING;
    }
  }

  /*
   * Tries to split a string command (e.g. 'mqtt 60' or 'show_csv') into
   * a CommandLine struct (function_name, argument_type and argument)
   */
  void parseCommand(const char *command, CommandLine &command_line) {
    if (strlen(command) == 0) {
      Serial.println(F("Received empty command"));
      command_line.argument_type = NONE;
      return;
    }

    char *first_space;
    first_space = strchr(command, ' ');

    if (first_space == NULL) {
      command_line.argument_type = NONE;
      strlcpy(command_line.function_name, command, MAX_COMMAND_SIZE);
      return;
    }

    strlcpy(command_line.function_name, command, first_space - command + 1);
    strlcpy(command_line.str_argument, first_space + 1, MAX_COMMAND_SIZE - (first_space - command) - 1);

    char *end;
    command_line.int_argument = strtol(command_line.str_argument, &end, 0); // Accepts 123 or 0xFF00FF

    if (*end) {
      command_line.argument_type = STRING;
    } else {
      command_line.argument_type = INT32;
    }
  }

  int compareCommandNames(const void *s1, const void *s2) {
    struct Command *c1 = (struct Command*) s1;
    struct Command *c2 = (struct Command*) s2;
    return strcmp(c1->name, c2->name);
  }

  void listAvailableCommands() {
    qsort(commands, commands_count, sizeof(commands[0]), compareCommandNames);
    for (uint8_t i = 0; i < commands_count; i++) {
      Serial.print(F("  "));
      Serial.print(commands[i].name);
      Serial.print(F(" "));
      Serial.print(commands[i].doc);
      Serial.println(F("."));
    }
  }

  /*
   * Saves bytes from Serial.read() until enter is pressed, and tries to run the corresponding command.
   *   http://www.gammon.com.au/serial
   */
  void processSerialInput(const byte input_byte) {
    static char input_line[MAX_COMMAND_SIZE];
    static unsigned int input_pos = 0;
    switch (input_byte) {
    case '\n': // end of text
      Serial.println();
      input_line[input_pos] = 0;
      execute(input_line);
      input_pos = 0;
      break;
    case '\r': // discard carriage return
      break;
    case '\b': // backspace
      if (input_pos > 0) {
        input_pos--;
        Serial.print(F("\b \b"));
      }
      break;
    default:
      if (input_pos == 0) {
        Serial.print(F("> "));
      }
      // keep adding if not full ... allow for terminating null byte
      if (input_pos < (MAX_COMMAND_SIZE - 1)) {
        input_line[input_pos++] = input_byte;
        Serial.print((char) input_byte);
      }
      break;
    }
  }

  void checkSerialInput() {
    while (Serial.available() > 0) {
      sensor_console::processSerialInput(Serial.read());
    }
  }

  /*
   * Tries to find the corresponding callback for a given command. Name and parameter type should fit.
   */
  void execute(const char *command_str) {
    CommandLine input;
    parseCommand(command_str, input);
    for (uint8_t i = 0; i < commands_count; i++) {
      if (!strcmp(input.function_name, commands[i].name) && input.argument_type == commands[i].parameter_type) {
        Serial.print(F("Calling : "));
        Serial.print(input.function_name);
        switch (input.argument_type) {
        case NONE:
          Serial.println(F("()"));
          commands[i].voidFunction();
          return;
        case INT32:
          Serial.print(F("("));
          Serial.print(input.int_argument);
          Serial.println(F(")"));
          commands[i].intFunction(input.int_argument);
          return;
        case STRING:
          Serial.print(F("('"));
          Serial.print(input.str_argument);
          Serial.println(F("')"));
          commands[i].strFunction(input.str_argument);
          return;
        }
      }
    }
    Serial.print(F("'"));
    Serial.print(command_str);
    Serial.println(F("' not supported. Available commands :"));
    listAvailableCommands();
  }
}
